/*
 *    author: Lloyd Dupont
 *    mail: lloyd@galador.net
 *    Copyright (C) 2000 Lloyd Dupont
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
//package utils;
import java.io.*;
import java.util.*;

/**
 * Provides the capability to read and write OpenStep style property
 * lists.  Arrays are delimited by parentheses with individual elements
 * separated by commas and optional spaces. <BR>
 * Hashtables (i.e. NSDictionary objects in OpenStep) are delimited by
 * curly braces, with key-value pairs separated by semicolons, the
 * keys and values separated by an equal sign. <BR>
 * Strings appears as plain text if they contain no whitespace, or
 * enclosed in straight quotation marks if they do. <BR>
 * Byte arrays (i.e. NSData objects in OpenStep) are delimited by
 * angled brackets and are encoded as hexadecimal digits. <BR>
 * Here's a short example of a text-format property list: <PRE>
 * {
 *     Title = "Star Wars";
 *     Director = "Lucas, George";
 *     Cast = (
 *         "Hamil, Mark",
 *         "Fisher, Carrier",
 *         "Ford, Harrison"
 *         );
 *     "Thumbnail Image" = <040b47f9 (many more sets of digits) 8484074e>
 * } </PRE>
 * <BR>
 * &nbsp; &nbsp; This object is a bit tolerant on error, bad separator
 * and such stuff, but rewrite PList correctly.
 *
 * @see #read
 * @see #write
 */
public class PList
{
  private static String BADCHAR   = "malformed plist ";
  private static String BADEND    = "unexpected end of plist";
  private static String BADOBJECT = "plist object is none of " +
        " Hastable, Vector, String or byte[]";

  public static void main(String[] args)
    throws Exception
  {
    if(args.length == 0)
      return;

    long l0 = System. currentTimeMillis();
    Object o = read( new FileReader(args[0]) );
    long l1 = System. currentTimeMillis();

    Writer w;
    if(args.length > 1)
      w = new FileWriter(args[1]);
    else
      w = new OutputStreamWriter(System.out);
    write(w, o);

    System.out.println("took " + (l1-l0) + " ms.");
  }

  /** write a PList (Hashtable,Vector,String or byte[]) in a Text stream */
  public static void write(Writer w, Object plist)
    throws IOException
  {
    writeobject(w, 0, plist);
    w.flush();
  }
  private static void writeobject(Writer w, int lvl, Object o)
    throws IOException
  {
    if(o instanceof Hashtable)
      writedict(w, lvl, (Hashtable) o);
    else if(o instanceof Vector)
      writearray(w, lvl, (Vector) o);
    else if(o instanceof String)
      writestring(w, (String) o);
    else if(o instanceof byte[])
      writebytes(w, (byte[]) o);
    else
      throw new IllegalArgumentException(BADOBJECT);
  }
  private static void printwhite(Writer w, int lvl)
    throws IOException
  {
    char[] buf = new char[2*lvl];
    for(int i=0; i<buf.length; i++)
      buf[i] = ' ';
    w.write(buf);
  }
  private static void writedict(Writer w, int lvl, Hashtable h)
    throws IOException
  {
    w.write("{\n");
    Enumeration keys = h.keys();
    while(keys.hasMoreElements()) {
      String s = (String) keys.nextElement();
      printwhite(w, lvl+1);
      writestring(w, s);
      w.write(" = ");
      writeobject(w, lvl+1, h.get(s));
      w.write(";\n");
    }
    printwhite(w, lvl);
    w.write("}");
  }
  private static void writearray(Writer w, int lvl, Vector v)
    throws IOException
  {
    w.write("(\n");
    for(int i=0, n=v.size(); i<n; i++) {
      printwhite(w, lvl+1);
      writeobject(w, lvl+1, v.elementAt(i));
      if(i != n-1)
        w.write(",\n");
      else
        w.write("\n");
    }
    printwhite(w, lvl);
    w.write(")");
  }
  private static void writestring(Writer w, String s)
    throws IOException
  {
    w.write('"');
    char[] c = s.toCharArray();
    for(int i=0; i<c.length; i++) {
      if(c[i] == '\\' || c[i] == '\'' || c[i] == '"')
        w.write('\\');
      w.write(c[i]);
    }
    w.write('"');
  }
  private static void writebytes(Writer w, byte[] b)
    throws IOException
  {
    w.write('<');
    for(int i=0; i<b.length; i++)
    {
      if(i%4==0 && i!=0)
        w.write(" ");
      w.write(intToHexa(b[i]));
    }
    w.write('>');
  }
  private static char[] intToHexa(byte b)
  {
    int n = b < 0 ? b+256 : b;
    char[] tab = new char[2];
    tab[0] = miniToChar(n/16);
    tab[1] = miniToChar(n%16);
    return tab;
  }
  private static char miniToChar(int c)
  {
    if(c < 10)
      return (char)(c + '0');
    else
      return (char)(c + 'A' - 10);
  }

  /** read a PList (Hashtable,Vector,String or byte[]) from a Text stream */
  public static Object read(Reader r)
    throws IOException
  {
    return readobject(new PushbackReader(r, 1));
  }
  private static Object readobject(PushbackReader r)
    throws IOException
  {
    int c = next(r);
    switch(c)
    {
      case -1:
        throw new IllegalArgumentException(BADEND);
      case '{':
        return readdict(r);
      case '(':
        return readarray(r);
      case '<':
        return readbytes(r);
      default:
        return readstring(r);
    }
  }
  private static String readstring(PushbackReader r)
    throws IOException
  {
    int first = r.read();
    if(first == -1)
      throw new IllegalArgumentException(BADEND);

    StringBuffer sb = new StringBuffer();
    if(first != '\'' && first != '"')
    {
      r.unread(first);
      first = 0;
    }

    int c;
    while(true)
    {
      c = r.read();
      if(c == '\\') {
        c = r.read();
        if(c == -1)
          throw new IllegalArgumentException(BADEND);
      }
      else if(first != 0 && c == first)
        break;
      else if(first == 0) {
        if(isWhite(c) || c == -1)
          break;
        if(breakString(c)) {
          r.unread(c);
          break;
        }
      }
      else if(c == -1)
        throw new IllegalArgumentException(BADEND);
      sb.append((char) c);
    }
    return sb.toString();
  }
  private static boolean breakString(int c)
  {
    boolean ret = c=='=' || c ==',' || c == ';'
      || c =='(' || c ==')' || c =='{' || c =='}'
      || c =='<' || c=='>';
    return ret;
  }
  private static Hashtable readdict(PushbackReader r)
    throws IOException
  {
    Hashtable h = new Hashtable();
    int c = r.read();
    if(c != '{')
      throw new IllegalArgumentException(BADCHAR + (char)c);
    while(true)
    {
      c = next(r);
      if(c == '}') {
        r.read();
        break;
      }
      String key = readstring(r);

      c = next(r);
      if(c != '=')
        throw new IllegalArgumentException(BADCHAR + (char)c);
      r.read();

      next(r);
      Object value = readobject(r);
      h.put(key, value);

      c = next(r);
      if(c == '}') {
        r.read();
        break;
      }
      if(c == ';' || c == ',') {
        r.read();
        continue;
      }
    }
    return h;
  }
  private static Vector readarray(PushbackReader r)
    throws IOException
  {
    Vector v = new Vector();
    int c = r.read();
    if(c != '(')
      throw new IllegalArgumentException(BADCHAR + (char)c);
    while(true)
    {
      c = next(r);
      if(c == ')') {
        r.read();
        break;
      }

      Object item = readobject(r);
      v.addElement(item);

      c = next(r);
      if(c == ')') {
        r.read();
        break;
      }
      if(c == ';' || c == ',') {
        r.read();
        continue;
      }
      // be tolerant
      // throw new IllegalArgumentException(BADCHAR + (char)c);
    }
    return v;
  }
  private static byte[] readbytes(PushbackReader r)
    throws IOException
  {
    ByteArrayOutputStream bos = new ByteArrayOutputStream();
    int c = r.read();
    if(c != '<')
      throw new IllegalArgumentException(BADCHAR + (char)c);

    char[] tab = new char[2];
    int read;
    while(true)
    {
      while(isWhite(read = r.read()));
      if(read == -1)
        throw new IllegalArgumentException(BADEND);
      if(read  == '>')
        break;
      tab[0] = (char) read;
      while(isWhite(read = r.read()));
      if(read == -1)
        throw new IllegalArgumentException(BADEND);
      tab[1] = (char) read;

      read = hexaToInt(tab[0]);
      read <<= 4;
      read |= hexaToInt(tab[1]);
      bos.write(read);
    }
    return bos.toByteArray();
  }
  private static int hexaToInt(char c)
  {
    int n;
    if('0' <= c && c <= '9')
      n = c - '0';
    else if('a' <= c && c <= 'f')
      n = c - 'a' + 10;
    else if('A' <= c && c <= 'F')
      n = c - 'A' + 10;
    else
      throw new IllegalArgumentException(BADCHAR + (char)c);
    return n;
  }
  private static int next(PushbackReader r)
    throws IOException
  {
    int c;
    while(isWhite( c = r.read() ));
    r.unread(c);
    return c;
  }
  private static boolean isWhite(int c)
  {
    boolean b = c != ' ' && c != '\t' && c != '\r' && c != '\n';
    return !b;
  }
}
